import os
import re
import sys
import numpy as np
from io import StringIO
from pathlib import Path
from pandas.io.formats.format import GenericArrayFormatter
from Bio import SeqIO
import sys
import statistics
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import gzip
from mimetypes import guess_type
from functools import partial

codon_table = {
    'GCT':'A', 'GCC':'A', 'GCA':'A', 'GCG':'A', 'CGT':'R', 'CGC':'R',   
    'CGA':'R', 'CGG':'R', 'AGA':'R', 'AGG':'R', 'TCT':'S', 'TCC':'S',
    'TCA':'S', 'TCG':'S', 'AGT':'S', 'AGC':'S', 'ATT':'I', 'ATC':'I',
    'ATA':'I', 'TTA':'L', 'TTG':'L', 'CTT':'L', 'CTC':'L', 'CTA':'L',
    'CTG':'L', 'GGT':'G', 'GGC':'G', 'GGA':'G', 'GGG':'G', 'GTT':'V',
    'GTC':'V', 'GTA':'V', 'GTG':'V', 'ACT':'T', 'ACC':'T', 'ACA':'T',
    'ACG':'T', 'CCT':'P', 'CCC':'P', 'CCA':'P', 'CCG':'P', 'AAT':'N',
    'AAC':'N', 'GAT':'D', 'GAC':'D', 'TGT':'C', 'TGC':'C', 'CAA':'Q',
    'CAG':'Q', 'GAA':'E', 'GAG':'E', 'CAT':'H', 'CAC':'H', 'AAA':'K',
    'AAG':'K', 'TTT':'F', 'TTC':'F', 'TAT':'Y', 'TAC':'Y', 'ATG':'M',
    'TGG':'W',
    'TAG':'STOP', 'TGA':'STOP', 'TAA':'STOP'
    }

def fasta_dict(file: Path):
    """
    Reads a file of FASTA entries and returns a dict where each key is a sequence ID.
    The value is another dict with two keys 'h' for header and 's' for sequence.  The
    header is all the other text after the id in the original FASTA header.  The
    sequence has all whitespace removed.  Obviously this should only be used on files
    where memory to load them isn't an issue.
    """
    seqs = dict()
    current_seq = ''
    current_id = None
    current_header = None
    
    for line in open(file):
        line = line.rstrip()
        m = re.search('>(\S+)\s*(.*)', line)
        if m:
            ## new residue line matched, purge the existing one, if not the first
            if current_id is not None:
                ## warn if it has already been found
                if current_id in seqs:
                    sys.stderr.write(f"WARN: Duplicate ID ({current_id}) found.  Only last one kept.\n")

                ## remove all whitespace and save
                current_seq = ''.join(current_seq.split())
                seqs[current_id] = {'h':current_header, 's':current_seq}
                    
            current_seq = ''
            current_id = m.group(1)
            current_header = m.group(2)
        else:
            current_seq += str(line)

    ## don't forget the last one
    current_seq = ''.join(current_seq.split())
    seqs[current_id] = {'h':current_header, 's':current_seq}

    return seqs

def fasta_sizes(file):
    """
    Reads a file of FASTA entries and returns a dict where each key is a sequence ID and
    the value is just the size of each sequence.  This is as almost as computationally
    intensive as fasta_dict_from_file() but takes less memory and is appropriate when you
    only care about the residue lengths.
    """
    seqs = dict()
    current_id = None
    
    for line in open(file):
        line = line.rstrip()
        m = re.search('>(\S+)\s*(.*)', line)
        if m:
            ## new residue line matched, set the new seq ID
            current_id = m.group(1)
            seqs[current_id] = 0
        else:
            seqs[current_id] += len(line)

    return seqs


def splitFileContents(f, delimiter, BLOCKSIZE=8192):
    """
    Same semantics as f.read().split(delimiter), but with memory usage
    determined by largest chunk rather than entire file size
    """
    remainder = StringIO()
    while True:
        block = f.read(BLOCKSIZE)
        if not isinstance(block, str):
            block = block.decode("utf-8")
        if not block:
            break
        parts = block.split(delimiter)
        remainder.write(parts[0])
        for part in parts[1:]:
            yield remainder.getvalue()
            remainder = StringIO()
            remainder.write(part)
    yield remainder.getvalue()


def get_genome_stats(infile, genome_size):
  """Calculate the statistics for an input fasta formatted genome file.
        infile should be a fasta formatted file
        genome_size should be in MBp
        Returns a dictionary of values for that genome.
  """
  
  # If infile ends in .gz us gzip.open, otherwise use standard open.
  encoding = guess_type(infile)[1]  # uses file extension
  _open = partial(gzip.open, mode='rt') if encoding == 'gzip' else open
  
  with _open(infile) as f:
    records = list(SeqIO.parse(f, "fasta"))
  
  genome_dict={}
  
  genome_dict['number_of_scaffolds'] = len(records)
  genome_dict['len_seq'] = [len(rec) for rec in records]
  genome_dict['scaffold_lengths'] = pd.DataFrame(genome_dict['len_seq'])
  genome_dict['total_size_scaffolds'] = sum(genome_dict['len_seq'])

  genome_dict['total_scaffold_length_percentage_genome_size'] = ((genome_dict['total_size_scaffolds']/(genome_size*1000000))*100)
    
  genome_dict['seq_greater_1k'] = sorted(i for i in genome_dict['len_seq'] if i>1000)
  genome_dict['seq_greater_10k'] = sorted(i for i in genome_dict['seq_greater_1k'] if i>10000)
  genome_dict['seq_greater_25k'] = sorted(i for i in genome_dict['seq_greater_10k'] if i>25000)
  genome_dict['seq_greater_100k'] = sorted(i for i in genome_dict['seq_greater_25k'] if i>100000)
  genome_dict['seq_greater_1M'] = sorted(i for i in genome_dict['seq_greater_100k'] if i>1000000)
  genome_dict['seq_greater_10M'] = sorted(i for i in genome_dict['seq_greater_1M'] if i>10000000)

  genome_dict['sorted_len'] = sorted(genome_dict['len_seq'], reverse=True)
  genome_dict['sum_sorted_length'] = sum(genome_dict['sorted_len'])
  genome_dict['half_length'] = genome_dict['sum_sorted_length']/2.0

  #calculates N50 and L50 values
  testSum = 0
  genome_dict['N50'] = 0
  genome_dict['L50'] = 0
  i = 0
  half_length = genome_dict['sum_sorted_length']/2.0

  for con in genome_dict['sorted_len']:
      testSum += con
      i += 1
      if half_length < testSum:
        genome_dict['N50'] = con
        genome_dict['L50'] = i
        break

  #calculates NG50 and LG50 values
  half_genome = (genome_size*1000000)/2.0
  testSumNG50 = 0
  genome_dict['NG50'] = 0
  genome_dict['LG50'] = 0
  i = 0
  for conNG50 in genome_dict['sorted_len']:
      testSumNG50 += conNG50
      i += 1
      if  half_genome < testSumNG50:
        genome_dict['NG50'] = conNG50
        genome_dict['LG50'] = i
        break

  #calculates A,C,G,T,N percentages
  genome_dict['counterAT'] = 0
  genome_dict['counterGC'] = 0

  genome_dict['counterN'] = 0
  for record in records:
      genome_dict['counterAT'] += record.seq.count('A') 
      genome_dict['counterAT'] += record.seq.count('T')
      genome_dict['counterGC'] += record.seq.count('G') 
      genome_dict['counterGC'] += record.seq.count('C') 
      genome_dict['counterN'] += record.seq.count('N') 

  return genome_dict

def write_output_stats(genome_stats, genome_size, num_longest, outputfile):
  """ Writes the genome stats to output file
        genome_stats is a dictionary of dictionaries produced by get_genome_stats
        genome_size should be in MBp
        num_longest is int of number of longest scaffold lengths to print
        outputfile is the file to print to
  """
  OUT = open(outputfile, 'w')

  OUT.write('Genome:')
  for genome in genome_stats:
    OUT.write(',')
    OUT.write(genome)
  OUT.write('\n')

  OUT.write('Number of scaffolds:')
  for genome in genome_stats:
    OUT.write(',')
    OUT.write(str(genome_stats[genome]['number_of_scaffolds']))
  OUT.write('\n')

  OUT.write('Total sum of scaffold lengths:')
  for genome in genome_stats:
    OUT.write(',')
    OUT.write(str(genome_stats[genome]['total_size_scaffolds']))
  OUT.write('\n')

  OUT.write('Percent of genome size:')
  for genome in genome_stats:
    OUT.write(',')
    OUT.write(str(genome_stats[genome]['total_scaffold_length_percentage_genome_size']))
  OUT.write('\n')

  OUT.write('Longest scaffold:')
  for genome in genome_stats:
    OUT.write(',')
    OUT.write(str(max(genome_stats[genome]['len_seq'])))
  OUT.write('\n')

  OUT.write('Shortest scaffold:')
  for genome in genome_stats:
    OUT.write(',')
    OUT.write(str(min(genome_stats[genome]['len_seq'])))
  OUT.write('\n\n')

  OUT.write('--------------------------------------------------------------------\n')
  OUT.write('\n')

  OUT.write('Total no. scaffolds over 1KBp:')
  for genome in genome_stats:
    OUT.write(',')
    OUT.write(str(len(genome_stats[genome]['seq_greater_1k'])))
  OUT.write('\n')

  OUT.write(f'Sum of scaffold lengths over 1KBp:')
  for genome in genome_stats:
    OUT.write(',')
    OUT.write(str(sum(genome_stats[genome]['seq_greater_1k'])))
  OUT.write('\n')

  OUT.write(f'Percent genome over 1KBp:')
  for genome in genome_stats:
    OUT.write(',')
    OUT.write(str(sum(genome_stats[genome]['seq_greater_1k'])/(genome_size*1000000)*100))
  OUT.write('\n')

  OUT.write('\n')
  OUT.write('Total no. scaffolds over 10KBp:')
  for genome in genome_stats:
    OUT.write(',')
    OUT.write(str(len(genome_stats[genome]['seq_greater_10k'])))
  OUT.write('\n')

  OUT.write(f'Sum of scaffold lengths over 10KBp:')
  for genome in genome_stats:
    OUT.write(',')
    OUT.write(str(sum(genome_stats[genome]['seq_greater_10k'])))
  OUT.write('\n')

  OUT.write(f'Percent genome over 10KBp:')
  for genome in genome_stats:
    OUT.write(',')
    OUT.write(str(sum(genome_stats[genome]['seq_greater_10k'])/(genome_size*1000000)*100))
  OUT.write('\n')

  OUT.write('\n')
  OUT.write('Total no. scaffolds over 25KBp:')
  for genome in genome_stats:
    OUT.write(',')
    OUT.write(str(len(genome_stats[genome]['seq_greater_25k'])))
  OUT.write('\n')

  OUT.write(f'Sum of scaffold lengths over 25KBp:')
  for genome in genome_stats:
    OUT.write(',')
    OUT.write(str(sum(genome_stats[genome]['seq_greater_25k'])))
  OUT.write('\n')

  OUT.write(f'Percent genome over 25KBp:')
  for genome in genome_stats:
    OUT.write(',')
    OUT.write(str(sum(genome_stats[genome]['seq_greater_25k'])/(genome_size*1000000)*100))
  OUT.write('\n')

  OUT.write('\n')
  OUT.write('Total no. scaffolds over 100KBp:')
  for genome in genome_stats:
    OUT.write(',')
    OUT.write(str(len(genome_stats[genome]['seq_greater_100k'])))
  OUT.write('\n')

  OUT.write(f'Sum of scaffold lengths over 100KBp:')
  for genome in genome_stats:
    OUT.write(',')
    OUT.write(str(sum(genome_stats[genome]['seq_greater_100k'])))
  OUT.write('\n')

  OUT.write(f'Percent genome over 100KBp:')
  for genome in genome_stats:
    OUT.write(',')
    OUT.write(str(sum(genome_stats[genome]['seq_greater_100k'])/(genome_size*1000000)*100))
  OUT.write('\n')

  OUT.write('\n')
  OUT.write('Total no. scaffolds over 1MBp:')
  for genome in genome_stats:
    OUT.write(',')
    OUT.write(str(len(genome_stats[genome]['seq_greater_1M'])))
  OUT.write('\n')

  OUT.write(f'Sum of scaffold lengths over 1MBp:')
  for genome in genome_stats:
    OUT.write(',')
    OUT.write(str(sum(genome_stats[genome]['seq_greater_1M'])))
  OUT.write('\n')

  OUT.write(f'Percent genome over 1MBp:')
  for genome in genome_stats:
    OUT.write(',')
    OUT.write(str(sum(genome_stats[genome]['seq_greater_1M'])/(genome_size*1000000)*100))
  OUT.write('\n')

  OUT.write('\n')
  OUT.write('Total no. scaffolds over 10MBp:')
  for genome in genome_stats:
    OUT.write(',')
    OUT.write(str(len(genome_stats[genome]['seq_greater_10M'])))
  OUT.write('\n')

  OUT.write(f'Sum of scaffold lengths over 10MBp:')
  for genome in genome_stats:
    OUT.write(',')
    OUT.write(str(sum(genome_stats[genome]['seq_greater_10M'])))
  OUT.write('\n')

  OUT.write(f'Percent genome over 10MBp:')
  for genome in genome_stats:
    OUT.write(',')
    OUT.write(str(sum(genome_stats[genome]['seq_greater_10M'])/(genome_size*1000000)*100))
  OUT.write('\n\n')
  OUT.write('--------------------------------------------------------------------\n')
  OUT.write('\n')
  
  OUT.write('N50:')
  for genome in genome_stats:
    OUT.write(',')
    OUT.write(str(genome_stats[genome]['N50']))
  OUT.write('\n')

  OUT.write('L50:')
  for genome in genome_stats:
    OUT.write(',')
    OUT.write(str(genome_stats[genome]['L50']))
  OUT.write('\n')

  OUT.write('NG50:')
  for genome in genome_stats:
    OUT.write(',')
    OUT.write(str(genome_stats[genome]['NG50']))
  OUT.write('\n')

  OUT.write('LG50:')
  for genome in genome_stats:
    OUT.write(',')
    OUT.write(str(genome_stats[genome]['LG50']))
  OUT.write('\n\n')
  OUT.write('--------------------------------------------------------------------\n')
  OUT.write('\n')

  OUT.write('%AT:')
  for genome in genome_stats:
    OUT.write(',')
    OUT.write(str((genome_stats[genome]['counterAT']/genome_stats[genome]['total_size_scaffolds'])*100))
  OUT.write('\n')

  OUT.write('%GC:')
  for genome in genome_stats:
    OUT.write(',')
    OUT.write(str((genome_stats[genome]['counterGC']/genome_stats[genome]['total_size_scaffolds'])*100))
  OUT.write('\n')

  OUT.write('%N:')
  for genome in genome_stats:
    OUT.write(',')
    OUT.write(str((genome_stats[genome]['counterN']/genome_stats[genome]['total_size_scaffolds'])*100))
  OUT.write('\n')

  OUT.write(f'\n')
  OUT.write('--------------------------------------------------------------------\n')
  OUT.write('\n')
  # Print longest n scaffold lengths
  OUT.write(f'Longest {num_longest} scaffolds:')
  OUT.write(f'')
  for genome in genome_stats:
    OUT.write(',')
    nlongest=(genome_stats[genome]['scaffold_lengths'].sort_values(0,ascending=False).head(n=num_longest))
    for index in nlongest.index:
      OUT.write(f'{nlongest[0][index]} ')

  OUT.write('\n')

  OUT.close()


def plot_scaffold_dist(genome_stats, outputfile):
  """Print a histogram of the scaffold length distributions
      genome_stats is a dictionary of dictionaries produced by get_genome_stats
      outputfile is the file to print graph to
  """
  for genome in genome_stats:
    plt.hist(genome_stats[genome]['len_seq'], label=genome, bins=30, log=True, alpha=0.5)

  plt.legend()

  plt.xlabel("Scaffold Length", fontsize=16)  
  plt.ylabel("Log count", fontsize=16)

  plt.savefig(outputfile)






import os, re, sys
import logging
import time
import typer
import math
import hashlib
import pandas as pd
import subprocess
from rich import print
from rich.progress import track
from typing import List, Union, Optional, NamedTuple
from openpyxl import load_workbook
from openpyxl.workbook.workbook import Workbook
from openpyxl.worksheet.worksheet import Worksheet
from collections import defaultdict
from collections import OrderedDict
from Bio.Seq import Seq
from Bio import SeqIO
from Bio.SeqRecord import SeqRecord
from pathlib import Path
from lxzbiotools.utils import codon_table, fasta_dict, get_genome_stats, write_output_stats, plot_scaffold_dist


# start=time.perf_counter()
app=typer.Typer(help="Xingze Li's bioinformatics analysis scripts. \n\n\n emali: lixingzee@gmail.com")

@app.command("m2ofa")
def to_one_line_fasta(inf: Path = typer.Argument(...,
                                                 help="input fasta file"),
                      outf: Path = typer.Argument(...,
                                                 help="output fasta file")):
    'Convert multi-line fasta to one-line fasta'
    inf = open(inf, 'r')
    outf = open(outf, 'w')
    db = {}
    for line in track(inf.readlines()):
        if line.startswith('>'):
            keys = line.strip()
            db[keys] = []
        else:
            db[keys].append(line.strip())
    for name, seq in db.items():
        outf.write(name + '\n')
        outf.write(''.join(seq) + '\n')
    inf.close()
    outf.close()

@app.command('extseq')
def extract_target_seq(inf: Path = typer.Argument(...,
                                                  help="input fasta file"), 
                       outf: Path = typer.Argument(...,
                                                   help="output result file"), 
                       ext: Path = typer.Argument(...,
                                                                    help="extract sequences name file")):
    'Extract sequences by sequence name or keyword'
    outf = open(outf,'w')
    dict = {}
    with open(inf, 'r') as fastaf:
        for line in fastaf:
            if line.startswith('>'):
                name = line
                dict[name] = ''
            else:
                dict[name] += line.replace('\n','') 
         
    with open(ext,'r') as listf:
        for row in listf:
            row = row.strip()
            for key in dict.keys(): 
                if row in key:
                    outf.write(key)
                    outf.write(dict[key] + '\n')
    outf.close()

@app.command()
def fq2fa(fq: Path = typer.Argument(...,
                                    help="input fastq file"),
          fa: Path = typer.Argument(...,
                                    help="output fasta file")):
    'Convert a fastq file to a fasta file'
    out=[]
    with open(fq,'r') as fin:
        c=0
        for line in fin:
            c+=1
            if (c%4) == 1:
                out.append('>' + line[1:])
            elif (c%4) == 2:
                out.append(line)
    with open(fa,'w') as fo:
        fo.write(''.join(out))
    return

@app.command()
def fa2fq(fa: Path = typer.Argument(...,
                                    help="input fasta file"),
          fq: Path = typer.Argument(...,
                                    help="output fastaq file")):
    'Convert a fasta file to a fastq file'
    seq=''
    out=[]
    n_reads=0
    with open(fa,'r') as fin:
        for line in track(fin):
            if line.startswith('>'):
                n_reads+=1
                if seq:
                    score='I'*len(seq)
                    out.append(seq+'\n+\n'+score+'\n')
                    seq=''
                out.append('@'+line[1:])
            else:
                seq+=line.strip()
    score='I'*len(seq)
    out.append(seq+'\n+\n'+score+'\n')
    with open(fq,'w') as fo:
        fo.write(''.join(out))
    return n_reads



@app.command()
def cds2pep(cds_input: Path  = typer.Argument(...,
                                              help="input coding sequence(CDS) file"),
            cds_output: Path = typer.Argument(...),
            pep_output: Path = typer.Argument(...,
                                              help="output pep file")):
    'Convert cds file to pep file'
    inf = open(cds_input, 'r')
    cds_outf = open(cds_output, 'w')        
    pep_outf = open(pep_output, 'w')
    db = {}
    pep = {}
    for line in track(inf.readlines()):
        if line.startswith('>'):
            keys = line.strip()
            db[keys] = []
        else:
            db[keys].append(line.strip())

    for name, seq in track(db.items()):
        cds_outf.write(name + '\n')
        cds_outf.write(''.join(seq) + '\n')
    cds_outf.close()

    with open(cds_output, "r") as co:
        for line in co.readlines():
            if line.startswith('>'):
                key = line.strip()
                pep[key] = []
            else:
                seq = line.strip()
                # translate one frame at a time
                prot = '' 
                for i in range(0, len(seq), 3):
                    codon = seq[i:i + 3]
                    if codon in codon_table:
                        if codon_table[codon] == 'STOP':
                            prot = prot + '*'
                        else: 
                            prot = prot + codon_table[codon]
                pep[key].append(prot)
        for name, seq in track(pep.items()):
            pep_outf.write(name + '\n')
            pep_outf.write(''.join(seq) + '\n')
    inf.close()
    pep_outf.close()



@app.command()
def excel2txt(input: Path,
              output: Path):
    'Convert excel file to txt file'
    df = pd.read_excel(input, header=None)		
    df.to_csv(output, header=None, sep='\t', index=False)	

@app.command('length')
def get_fasta_len(fasta: Path,
                  lenf: Path):
    'Get the length of each sequence'
    fasta_dict = OrderedDict()
    gene_count = {}
    handle = open(fasta, "r")
    len_file = open(lenf, "w")
    active_sequence_name = ""
    for line in handle:
        line = line.strip()
        if not line:
            continue
        if line.startswith(">"): 
            active_sequence_name = line[1:]
            active_sequence_name = active_sequence_name.split(" ")[0]
        if active_sequence_name not in fasta_dict:
            fasta_dict[active_sequence_name] = 0
            continue
        sequence = line
        fasta_dict[active_sequence_name] += len(sequence)
    for chrom,lens in fasta_dict.items():
        print("{chrom}\t{lens}\t{count}".format(\
                chrom=chrom,lens=lens,count=gene_count.get(chrom,0)), file=len_file)
    len_file.close()    
    handle.close()
    return fasta_dict


@app.command('gff')
def parse_gff(gff: Path,
              sample_gff_file: Path):
    'Simplify gff3 file for WGD event analysis'
    gene_count = {}
    gene_dict = OrderedDict()
    tx_pos_dict = defaultdict(list)
    CDS_dict = defaultdict(list)
    handle = open(gff, "r")
    gff_file = open(sample_gff_file, "w")    
    for line in track(handle):
        if line.startswith("#"):
            continue
        content = line.split("\t")
        if len(content) <= 8:
            continue
        #print(content)
        if content[2] == "transcript" or content[2] == "mRNA":

            # use regual expression to extract the gene ID
            # match the pattern ID=xxxxx; or ID=xxxxx
            
            tx_id = re.search(r'ID=(.*?)[;\n]',content[8]).group(1)
            tx_parent = re.search(r'Parent=(.*?)[;\n]',content[8]).group(1)
            tx_parent = tx_parent.strip() # remove the 'r' and '\n'

            # if the parent of transcript is not in the gene_dict, create it rather than append
            if tx_parent in gene_dict:
                gene_dict[tx_parent].append(tx_id)
            else:
                gene_dict[tx_parent] = [tx_id]
            tx_pos_dict[tx_id] = [content[0],content[3], content[4], content[6] ]
        # GFF must have CDS feature
        if content[2] == 'CDS':
            width = int(content[4]) - int(content[3])
            CDS_parent = re.search(r'Parent=(.*?)[;\n]',content[8]).group(1)
            CDS_parent = CDS_parent.strip() # strip the '\r' and '\n'
            CDS_dict[CDS_parent].append(width)
    
    for gene, txs in gene_dict.items():
        tmp = 0
        for tx in txs:
            tx_len = sum(CDS_dict[tx])
            if tx_len > tmp:
                lst_tx = tx
                tmp = tx_len
        tx_chrom = tx_pos_dict[lst_tx][0]
        if tx_chrom not in gene_count:
            gene_count[tx_chrom] = 1
        else:
            gene_count[tx_chrom] += 1
        tx_start = tx_pos_dict[lst_tx][1]
        tx_end   = tx_pos_dict[lst_tx][2]
        tx_strand = tx_pos_dict[lst_tx][3]

        print(f"{tx_chrom}\t{gene}\t{tx_start}\t{tx_end}\t{tx_strand}\t{gene_count[tx_chrom]}\t{lst_tx}", file=gff_file )
    gff_file.close()
    handle.close()
    # return [gene_dict, tx_pos_dict, CDS_dict]




@app.command('parallel')
def parallel_run(cmd_path, 
                 thread, 
                 stdout=None, 
                 stderr=None, 
                 parallel="parallel", 
                 fg=False):
    'Parallelized running tasks'
    thread=int(thread)
    if stdout == None:
        stdout = "%s.log" % cmd_path
    if stderr == None:
        stderr = "%s.err" % cmd_path
    if thread >1:
        cmd = "nohup %s -j %s ::: < %s > %s 2> %s " % (parallel, thread, cmd_path, stdout, stderr)
    else:
        cmd = "nohup bash %s > %s 2> %s " %(cmd_path, stdout, stderr)
    if fg==False:
        cmd +=" &"
    sys.stdout.write("CMD: %s \n" % cmd)
    subprocess.check_call(cmd, shell=True, executable='/bin/bash')



@app.command()
def gfa2fa(gfa_file: Path,
           fa_file: Path):
    'Convert gfa file to fasta file'
    input = open(gfa_file, 'r').readlines()
    output = open(fa_file, 'w')
    print(f"Converting GFA {input} --> FASTA {output}...")
    num_seqs = 0
    for i in track(len(input)):
        line = input[i].strip('\n').split()
        if line[0] == 'S':
            num_seqs += 1
            simple_seq = Seq(line[2])
            simple_seq_r = SeqRecord(simple_seq, id=line[1])
            SeqIO.write(simple_seq_r, output, "fasta")
    output.close()
    print(f"FASTA of {num_seqs} sequences created at {output}.")

@app.command('rds')
def remove_duplicate_sequences(fasta_file:Path,
                               output_file:Path):
    'Read a multi-FASTA file sequence and remove duplicates (by MD5 hash)'
    ## output will either be a file or STDOUT
    '''
    Input: 
        - A (multi-)FASTA file.  
    Output:
        - Output will be multi-FASTA data printed to STDOUT or a file
    '''
    fout = sys.stdout
    if output_file is not None:
        fout = open(output_file, 'wt')
    
    seqs = fasta_dict(fasta_file)
    found = dict()

    m = hashlib.md5()

    for seq_id in track(seqs):
        seq = seqs[seq_id]['s']
        m.update(seq.encode())
        md5sum = m.hexdigest()
        
        ## write this sequence, 60bp per line
        if md5sum not in found:
            fout.write(f">{seq_id} {seqs[seq_id]['h']}\n")
            for i in range(0, len(seq), 60):
                fout.write(seq[i : i + 60] + "\n")
            found[md5sum] = 1


@app.command('genstats')
def genome_stats(infile: List[str] = typer.Option(...,"-i","--infile",
                                                help="input genome fasta"),
        genome_size: int = typer.Option(...,"-s","--size",
                                                help="input your genome size (MBp)"),
        outprefix: str = typer.Option("results","-o","--outprefix",
                                            help="specify output file prefix"),
        num_longest: int = typer.Option("10","-n","--nums",
                                          help="print the length of the longest scaffolds")):
    'single or multiple genome information statistics'
    
    print(f"Got {len(infile)} genome(s) to compare.")
        
    # Genome stats are stored in a dict of dicts, one for each genome.
    genomes_dict = {}
    append = 1

    # For each genome get the stats.
    for genome in infile:
        genome_name = os.path.split(genome)[1]

        if genome_name in genomes_dict:
            print(f"Multiple genome assemblies found with name {genome_name}, appending numbers to subsequent genomes")
            genome_name = genome_name + "_" + str(append)
            append += 1

        print(f"Getting statistics for {genome_name}.")
        genomes_dict[genome_name] = get_genome_stats(genome, genome_size)
        print(f"Found {(genomes_dict[genome_name]['number_of_scaffolds']):,} contigs for {genome_name}.")

    print(f"Done getting stats for {len(infile)} genomes. \nSummarizing data.")
    
    out_fasta = outprefix + '.csv'
    write_output_stats(genomes_dict, genome_size, num_longest, out_fasta)

    out_plots = outprefix + '.pdf'
    plot_scaffold_dist(genomes_dict, out_plots)

@app.command()
def movefile(file_dir: Path = typer.Argument(...,
                                    help="input your target files directory"),
             output_dir: Path = typer.Argument(...,
                                    help="output directory name"),
             dir_nums: int = typer.Argument(...,
                                    help="number of folders you want to create "),
             prefix: str = typer.Argument("",
                                    help="Folder prefix name, defaults to 1,2,3... 100")):
    'Randomly allocate files to a specified number of folders'
    file_dir = os.path.abspath(file_dir)
    folderPath = os.path.abspath(output_dir)
    sort_folder_number = [x for x in range(1,dir_nums+1)]
    for num in sort_folder_number:
        addfolder = prefix + str(num)
        new_folder_path = os.path.join(folderPath,addfolder) #new_folder_path is ‘folderPath\number'
        if not os.path.exists(new_folder_path):
            os.makedirs(new_folder_path)
            print("new a floder named "+str(num)+'at the path of '+ new_folder_path)

    #the taget file list
    file_list = os.listdir(file_dir)
    folderNumber = 0
    print('there are '+str(len(file_list))+' files at the path of '+file_dir)
    for i in range(0,len(file_list)):
        old_file_path = os.path.join(file_dir,file_list[i])
        if os.path.isdir(old_file_path):
            '''if the path is a folder,program will pass it'''
            print('file does not exist ,path=' + old_file_path+' it is a dir' )
            pass
        elif not os.path.exists(old_file_path):
            '''if the path does not exist,program will pass it'''
            print('file does not exist ,path='+old_file_path)
            pass
        else:
            '''define the number,it decides how many files each directory process'''
            number = math.ceil(len(file_list)/dir_nums)
            if(i%number == 0):
                folderNumber +=1
                addfile = prefix + str(folderNumber)
            new_file_path = os.path.join(folderPath,'%s'%(addfile))
            print(new_file_path)
            if not os.path.exists(new_file_path):
                print('not exist path:'+new_file_path)
                break
            shutil.move(old_file_path,new_file_path)
            print('success move file from '+ old_file_path +' to '+new_file_path)
    print('If you have any questions, please contact [bold magenta]lixingzee@gmail.com[/bold magenta]')
    print('[bold cyan]Thank you for using.[/bold cyan]')

    

# end=time.perf_counter()
# print('Running time: %s Seconds'%(end-start))   

if __name__ == "__main__":
    app()





# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['lxzbiotools']

package_data = \
{'': ['*']}

install_requires = \
['biopython>=1.79,<2.0',
 'matplotlib>=3.6.0,<4.0.0',
 'openpyxl>=3.0.10,<4.0.0',
 'pandas>=1.4.4,<2.0.0',
 'pathlib>=1.0.1,<2.0.0',
 'rich>=12.5.1,<13.0.0',
 'typer>=0.6.1,<0.7.0']

entry_points = \
{'console_scripts': ['lxzbiotools = lxzbiotools.lxzbiotools:app',
                     'utils = lxzbiotools.utils:app']}

setup_kwargs = {
    'name': 'lxzbiotools',
    'version': '0.5.4',
    'description': "Xingze Li's Bioinformatics Analysis Tools",
    'long_description': '## Project description\n\n**lxzbiotools** is a bioinformatics data processing tools.\n\n## Features\n\n+ `cds2pep`         Convert cds file to pep \n+ `excel2txt`       Convert file format form excel to txt\n+ `fa2fq`           Change fasta file to fastq file\n+ `fq2fa`           Change fastq file to fasta file\n+ `genstats`        One or more genome informatics statistics\n+ `gfa2fa`          Convert gfa file to fasta file\n+ `gff`             Simplify gff3 file for WGD event analysis\n+ `length`          Get the length of each sequences\n+ `rds`             Read a multi-FASTA file sequence and remove duplicates\n+ `parallel`        Parallelized running tasks\n+ `extseq`          Extract sequences by sequence name or keyword \n+ `m2ofa`           Convert multi-line fasta to one-line fasta           \n+ `movefile`        Randomly allocate files to a specified number of folders \n\n## QuickStart\n\n\n### Install\n\n```\npip3 install lxzbiotools\n```\n\n### Update\n\n```\npip3 install -U lxzbiotools\n```\n\n### Use\n\n```bash\n$ lxzbiotools --help\n\n Usage: lxzbiotools.py [OPTIONS] COMMAND [ARGS]...\n```\n\n\n## Version update content\n+ *2022-09-22* version **0.5.4** add **movefile** (Randomly allocate files to a specified number of folders)\n\n## Bug report \n\n+ Issues and bugs report to **lixingzee@gmail.com**\n\n+ github: *https://github.com/lxingze/lxzbiotools*\n\n',
    'author': 'Xingze_Li',
    'author_email': 'lixingzee@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
